import { ec as EC } from "elliptic";
import { encodeAbiParameters, encodeFunctionData, zeroAddress, } from "viem";
import { decodeErrorResult, publicKeyToAddress } from "viem/utils";
import { AUTHORIZER_ABI, ENTRYPOINT_ABI, ERC20_ABI, SMART_ACCOUNT_V1_ABI, SMART_ACCOUNT_V1_FACTORY_ABI, UNISWAP_V3_FACTORY_ABI, UNISWAP_V3_SWAP_ROUTER_ABI, } from "./abi";
import { DEX_CHAINS, ENTRYPOINT_ADDRESS, SMART_ACCOUNT_V1_CHAINS, } from "./addresses";
export class SmartAccountV1Provider {
    constructor(publicClient, authorizer, factory, dex) {
        Object.defineProperty(this, "ecdsa", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new EC("secp256k1")
        });
        Object.defineProperty(this, "publicClient", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "authorizer", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "factory", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "dex", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        this.publicClient = publicClient;
        const dexChain = publicClient.chain && DEX_CHAINS[publicClient.chain.id];
        this.dex = dex || dexChain;
        const smartAccountChain = publicClient.chain && SMART_ACCOUNT_V1_CHAINS[publicClient.chain.id];
        this.authorizer =
            authorizer || (smartAccountChain && smartAccountChain.authorizer);
        this.factory =
            factory || (smartAccountChain && smartAccountChain.smartAccountFactoryV1);
    }
    setDex(dex) {
        this.dex = dex;
    }
    setAuthorizer(authorizer) {
        this.authorizer = authorizer;
    }
    setFactory(factory) {
        this.factory = factory;
    }
    async createSessionKeyRequest(smartAccount, sessionNonce) {
        if (!this.authorizer) {
            throw new Error("Must provider authorizer contract");
        }
        const sessionKeyPair = this.ecdsa.genKeyPair();
        const sessionAddress = publicKeyToAddress(`0x${sessionKeyPair.getPublic("hex")}`);
        const [_fields, name, version, chainId, verifyingContract] = await this.publicClient.readContract({
            abi: AUTHORIZER_ABI,
            address: this.authorizer,
            functionName: "eip712Domain",
        });
        const types = {
            Permit: [
                {
                    name: "smartAccount",
                    type: "address",
                },
                {
                    name: "sessionAddress",
                    type: "address",
                },
                {
                    name: "sessionNonce",
                    type: "uint256",
                },
            ],
        };
        const request = {
            domain: {
                name,
                version,
                chainId: Number(chainId),
                verifyingContract,
            },
            types,
            primaryType: "Permit",
            message: {
                smartAccount,
                sessionAddress,
                sessionNonce,
            },
        };
        return {
            session: {
                privateKey: `0x${sessionKeyPair.getPrivate("hex")}`,
                address: sessionAddress,
            },
            request,
        };
    }
    aggregateClientSignatures(ownerSignature, sessionSignature, sessionNonce) {
        const ownerSig = ownerSignature.slice(2);
        const sessionSig = sessionSignature.slice(2);
        const nonce = encodeAbiParameters([
            {
                type: "uint256",
            },
        ], [sessionNonce]).slice(2);
        return `0x${ownerSig + sessionSig + nonce}`;
    }
    async createSwapRequest(createSwapRequestInput) {
        if (!this.dex) {
            throw new Error("Must provide dex contracts");
        }
        if (!this.factory) {
            throw new Error("Must provide smartAccountV1Factory contract");
        }
        if (!createSwapRequestInput.smartAccount &&
            !createSwapRequestInput.initSmartAccountInput) {
            throw new Error("Must provide either `smartAccount` or `initSmartAccountInput`");
        }
        const weth = this.dex.weth;
        const dexName = createSwapRequestInput.dex;
        const dex = this.dex[dexName];
        if (!dex) {
            throw new Error(`${dexName} not supported`);
        }
        let smartAccount = createSwapRequestInput.smartAccount;
        let initCode = "0x";
        const initSmartAccountInput = createSwapRequestInput.initSmartAccountInput;
        if (initSmartAccountInput) {
            const { owner, salt } = initSmartAccountInput;
            const address = await this.getSmartAccountAddress(owner, salt);
            const bytecode = await this.publicClient.getBytecode({ address });
            if (bytecode && bytecode !== "0x") {
                throw new Error("Smart account already exists");
            }
            const initCodeCallData = encodeFunctionData({
                abi: SMART_ACCOUNT_V1_FACTORY_ABI,
                functionName: "createAccount",
                args: [owner, salt],
            });
            initCode = (this.factory.toLowerCase() +
                initCodeCallData.slice(2));
            smartAccount = address.toLowerCase();
        }
        if (!smartAccount) {
            throw new Error("Smart account not found");
        }
        const { tokenIn, tokenOut, fee, recipient, deadline, amountIn, amountOutMinimum, } = createSwapRequestInput.swapInput;
        if (tokenIn.toLowerCase() === tokenOut.toLowerCase()) {
            throw new Error("TokenIn and TokenOut must be different");
        }
        let swapRouterCalldata;
        if (tokenOut.toLowerCase() == weth.toLowerCase()) {
            const exactInputSingleCallData = encodeFunctionData({
                abi: UNISWAP_V3_SWAP_ROUTER_ABI,
                functionName: "exactInputSingle",
                args: [
                    {
                        tokenIn,
                        tokenOut,
                        fee,
                        recipient: zeroAddress,
                        deadline,
                        amountIn,
                        amountOutMinimum,
                        sqrtPriceLimitX96: BigInt(0),
                    },
                ],
            });
            const unwrapWeth9CallData = encodeFunctionData({
                abi: UNISWAP_V3_SWAP_ROUTER_ABI,
                functionName: "unwrapWETH9",
                args: [amountOutMinimum, recipient || smartAccount],
            });
            const multicallCallData = encodeFunctionData({
                abi: UNISWAP_V3_SWAP_ROUTER_ABI,
                functionName: "multicall",
                args: [[exactInputSingleCallData, unwrapWeth9CallData]],
            });
            swapRouterCalldata = multicallCallData;
        }
        else {
            const exactInputSingleCallData = encodeFunctionData({
                abi: UNISWAP_V3_SWAP_ROUTER_ABI,
                functionName: "exactInputSingle",
                args: [
                    {
                        tokenIn,
                        tokenOut,
                        fee,
                        recipient: recipient || smartAccount,
                        deadline,
                        amountIn,
                        amountOutMinimum,
                        sqrtPriceLimitX96: BigInt(0),
                    },
                ],
            });
            swapRouterCalldata = exactInputSingleCallData;
        }
        const dest = [];
        const value = [];
        const func = [];
        if (tokenIn.toLowerCase() === weth.toLowerCase()) {
            dest.push(dex.swapRouter);
            value.push(amountIn);
            func.push(swapRouterCalldata);
        }
        else {
            const approveCallData = encodeFunctionData({
                abi: ERC20_ABI,
                functionName: "approve",
                args: [dex.swapRouter, amountIn],
            });
            dest.push(tokenIn, dex.swapRouter);
            func.push(approveCallData, swapRouterCalldata);
        }
        const executeBatchCallData = encodeFunctionData({
            abi: SMART_ACCOUNT_V1_ABI,
            functionName: "executeBatch",
            args: [dest, value, func],
        });
        const suffixId = createSwapRequestInput.orderSeparatorId || 0;
        const nonce = await this.getNonceForSmartAccount(smartAccount, dex.factory, tokenIn, tokenOut, fee, suffixId);
        const userOpWithoutSign = {
            sender: smartAccount,
            nonce,
            initCode,
            callData: executeBatchCallData,
            callGasLimit: 0n,
            verificationGasLimit: 0n,
            preVerificationGas: 0n,
            maxFeePerGas: 0n,
            maxPriorityFeePerGas: 0n,
            paymasterAndData: "0x",
            signature: "0x",
        };
        userOpWithoutSign.callGasLimit = await this.calculateExecuteBatchGasLimit(smartAccount, dest, value, func);
        userOpWithoutSign.verificationGasLimit =
            await this.calculateVerificationGasLimitV1(userOpWithoutSign);
        if (!createSwapRequestInput.gasless) {
            userOpWithoutSign.maxFeePerGas = BigInt(30e9);
            userOpWithoutSign.maxPriorityFeePerGas = BigInt(2e9);
        }
        const userOpHash = await this.publicClient.readContract({
            abi: ENTRYPOINT_ABI,
            address: ENTRYPOINT_ADDRESS,
            functionName: "getUserOpHash",
            args: [userOpWithoutSign],
        });
        return { smartAccount, userOpHash, request: userOpWithoutSign };
    }
    async getSmartAccountAddress(owner, salt) {
        if (!this.factory) {
            throw new Error("Factory not supported");
        }
        const address = await this.publicClient.readContract({
            abi: SMART_ACCOUNT_V1_FACTORY_ABI,
            address: this.factory,
            functionName: "getAddress",
            args: [owner, salt],
        });
        return address;
    }
    async getNonceForSmartAccount(smartAccount, dexFactory, tokenIn, tokenOut, fee, suffixId) {
        const poolAddr = await this.publicClient.readContract({
            abi: UNISWAP_V3_FACTORY_ABI,
            address: dexFactory,
            functionName: "getPool",
            args: [tokenIn, tokenOut, fee],
        });
        const formattedPoolAddr = poolAddr.toLowerCase();
        const keyPrefix = BigInt(formattedPoolAddr) << BigInt(32);
        if (!(suffixId < 2 ** 31)) {
            throw new Error("Suffix id must be less than 2^31");
        }
        const zeroForOne = tokenIn.toLowerCase() < tokenOut.toLowerCase() ? 1n : 0n;
        const keySuffix = (zeroForOne << 31n) | BigInt(suffixId);
        const key = keyPrefix | keySuffix;
        const nonce = await this.publicClient.readContract({
            abi: ENTRYPOINT_ABI,
            address: ENTRYPOINT_ADDRESS,
            functionName: "getNonce",
            args: [smartAccount, key],
        });
        return nonce;
    }
    async calculateVerificationGasLimitV1(userOp) {
        const maxGas = 30000000n;
        const mockSignature = "0xfffffffffffffffffffffffffffffff0000000000000000000000000000000007aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa1c";
        const authorizerSignature = mockSignature.slice(2);
        const ownerSignature = mockSignature.slice(2);
        const sessionSignature = mockSignature.slice(2);
        const nonce = "".padEnd(64, "0");
        const data = encodeFunctionData({
            abi: ENTRYPOINT_ABI,
            functionName: "simulateValidation",
            args: [
                {
                    ...userOp,
                    verificationGasLimit: maxGas,
                    signature: `0x${authorizerSignature + ownerSignature + sessionSignature + nonce}`,
                },
            ],
        });
        const rpcUrl = this.publicClient.transport.url;
        if (!rpcUrl) {
            throw new Error("RPC URL not found");
        }
        const simulationRes = await fetch(rpcUrl, {
            method: "POST",
            headers: {
                "Content-Type": "application/json",
            },
            body: JSON.stringify({
                jsonrpc: "2.0",
                method: "eth_call",
                params: [
                    {
                        from: ENTRYPOINT_ADDRESS,
                        to: ENTRYPOINT_ADDRESS,
                        data,
                    },
                    "pending",
                ],
                id: 1,
            }),
        });
        const encodedData = await simulationRes.json();
        const result = decodeErrorResult({
            abi: ENTRYPOINT_ABI,
            data: encodedData.error.data,
        });
        if (result.errorName === "ValidationResult") {
            const returnInfo = result.args[0];
            const { preOpGas } = returnInfo;
            return preOpGas;
        }
        throw new Error(`Estimate verification gas: ${result.errorName}(${result.args})`);
    }
    async calculateExecuteBatchGasLimit(smartAccount, dest, value, func) {
        try {
            const estimatedGas = await this.publicClient.estimateContractGas({
                abi: SMART_ACCOUNT_V1_ABI,
                address: smartAccount,
                functionName: "executeBatch",
                args: [dest, value, func],
                account: ENTRYPOINT_ADDRESS,
            });
            const baseGas = BigInt(21000);
            if (estimatedGas <= 50000) {
                return BigInt(200000);
            }
            return estimatedGas - baseGas;
        }
        catch (error) {
            return BigInt(200000);
        }
    }
}
//# sourceMappingURL=smartAccountV1Provider.js.map